from typing import Iterable, List

from pcs.common.interface.dto import to_dict
from pcs.common.resource_agent_dto import ResourceAgentActionDto
from pcs.lib.resource_agent import ResourceAgentAction, ResourceAgentMetadata
from pcs.lib.cib.resource.const import OPERATION_ATTRIBUTES
from pcs.lib.cib.resource.types import ResourceOperationIn, ResourceOperationOut


# Operation monitor is always required, even if creating default actions was
# not requested or a resource/stonith agent does not specify it. See
# https://clusterlabs.org/pacemaker/doc/2.1/Pacemaker_Explained/html/resources.html#monitoring-resources-for-failure
_NECESSARY_OPERATIONS = {"monitor"}

_DEFAULT_INTERVALS = {"monitor": "60s"}


def get_default_operation_interval(operation_name: str) -> str:
    """
    Return default interval for given operation_name

    operation_name -- name of the operation to get the default interval for
    """
    return _DEFAULT_INTERVALS.get(operation_name, "0s")


def complete_operations_options(
    raw_operation_list: Iterable[ResourceOperationIn],
) -> List[ResourceOperationOut]:
    """
    Return a new list of operations with all necessary keys defined

    raw_operation_list -- entered operations
    """
    operation_list = []
    for raw_operation in raw_operation_list:
        operation = dict(raw_operation)
        if operation.get("interval") is None:
            operation["interval"] = get_default_operation_interval(
                str(operation["name"])
            )
        operation_list.append(operation)
    return operation_list


def get_default_operations(
    agent_metadata: ResourceAgentMetadata,
    necessary_only: bool = False,
    keep_extra_keys: bool = False,
) -> List[ResourceOperationOut]:
    """
    Return operations which should be put to a CIB resource on its creation
    """
    is_default_operation = (
        _is_default_operation_stonith
        if agent_metadata.name.is_stonith
        else _is_default_operation_resource
    )

    # filter actions
    action_list = [
        action
        for action in agent_metadata.actions
        if (necessary_only and action.name in _NECESSARY_OPERATIONS)
        or (not necessary_only and is_default_operation(action))
    ]

    # add necessary actions if they are missing
    defined_operation_names = frozenset(op.name for op in action_list)
    for op_name in _NECESSARY_OPERATIONS:
        if op_name not in defined_operation_names:
            action_list.append(
                ResourceAgentAction(
                    op_name, None, None, None, None, None, False, False
                )
            )

    # transform actions to operation definitions
    return [
        action_to_operation(action.to_dto(), keep_extra_keys=keep_extra_keys)
        for action in action_list
    ]


def action_to_operation(
    action: ResourceAgentActionDto, keep_extra_keys: bool = False
) -> ResourceOperationOut:
    """
    Transform agent action data to CIB operation data
    """
    # This function bridges new agent framework, which provides data in
    # dataclasses, to old resource create code and transforms new data
    # structures to a format expected by the old code. When resource create is
    # overhauled, this fuction is expected to be removed.
    operation = {}
    for key, value in to_dict(action).items():
        if key == "depth":
            # "None" values are not put to CIB, so this keeps the key in place
            # while making sure it's not put in CIB. I'm not sure why depth ==
            # 0 is treated like this, but I keep it in place so the behavior is
            # the same as it has been for a long time. If pcs starts using
            # depth / OCF_CHECK_LEVEL or there is other demand for it, consider
            # changing this so value of "0" is put in CIB.
            operation["OCF_CHECK_LEVEL"] = None if value == "0" else value
        elif key == "start_delay":
            operation["start-delay"] = value
        elif key in OPERATION_ATTRIBUTES or keep_extra_keys:
            operation[key] = value
    return operation


def _is_default_operation_resource(action: ResourceAgentAction) -> bool:
    # Copy all actions to the CIB even those not defined in the OCF standard
    # or pacemaker. This way even custom actions defined in a resource agent
    # will be copied to the CIB and run by pacemaker if they specify
    # an interval. See https://github.com/ClusterLabs/pcs/issues/132
    return action.name not in {
        # one-time action, not meant to be processed by pacemaker
        "meta-data",
        # deprecated alias of monitor
        "status",
        # one-time action, not meant to be processed by pacemaker
        "validate-all",
    }


def _is_default_operation_stonith(action: ResourceAgentAction) -> bool:
    return action.name == "monitor"



from typing import Any, Dict, Iterable, List, Optional

from pcs.common.interface.dto import to_dict
from pcs.common.reports import ReportItemSeverity, ReportProcessor
from pcs.lib.cib.resource.agent import (
    action_to_operation,
    complete_operations_options,
    get_default_operations,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import (
    find_one_resource_agent_by_type,
    list_resource_agents,
    list_resource_agents_ocf_providers,
    list_resource_agents_standards,
    list_resource_agents_standards_and_providers,
    ResourceAgentError,
    resource_agent_error_to_report_item,
    ResourceAgentFacadeFactory,
    ResourceAgentMetadata,
    split_resource_agent_name,
)
from pcs.lib.resource_agent.name import name_to_void_metadata


def list_standards(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents standards (ocf, lsb, ... ) on the local host
    """
    return [
        standard
        for standard in list_resource_agents_standards(lib_env.cmd_runner())
        if standard != "stonith"
    ]


def list_ocf_providers(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents ocf providers on the local host
    """
    return list_resource_agents_ocf_providers(lib_env.cmd_runner())


def list_agents_for_standard_and_provider(
    lib_env: LibraryEnvironment, standard_provider: Optional[str] = None
) -> List[str]:
    """
    List resource agents for specified standard on the local host

    standard_provider -- standard[:provider], e.g. None, ocf, ocf:pacemaker
    """
    if standard_provider:
        std_prov_list = [
            standard_provider[:-1]
            if standard_provider[-1] == ":"
            else standard_provider
        ]
    else:
        std_prov_list = list_resource_agents_standards_and_providers(
            lib_env.cmd_runner()
        )
    agents = []
    for std_prov in std_prov_list:
        if std_prov == "stonith":
            continue
        agents += list_resource_agents(lib_env.cmd_runner(), std_prov)
    return sorted(agents, key=str.lower)


# TODO return a list of DTOs
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all resource agents on the local host, optionally filtered and
        described

    describe -- load and return agents' metadata as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()

    # list agents for all standards and providers
    agent_names = []
    for std_prov in list_resource_agents_standards_and_providers(runner):
        if std_prov == "stonith":
            continue
        agent_names += [
            f"{std_prov}:{agent}"
            for agent in list_resource_agents(runner, std_prov)
        ]
    return _complete_agent_list(
        runner,
        lib_env.report_processor,
        sorted(agent_names, key=str.lower),
        describe,
        search,
    )


# backward compatibility layer - export agent metadata in the legacy format
def _agent_metadata_to_dict(
    agent: ResourceAgentMetadata, describe: bool = False
) -> Dict[str, str]:
    agent_dto = agent.to_dto()
    agent_dict = to_dict(agent_dto)
    agent_dict["actions"] = [
        action_to_operation(action, keep_extra_keys=True)
        for action in agent_dto.actions
    ]
    agent_dict["default_actions"] = (
        complete_operations_options(
            get_default_operations(agent, keep_extra_keys=True)
        )
        if describe
        else []
    )
    return agent_dict


def _complete_agent_list(
    runner: CommandRunner,
    report_processor: ReportProcessor,
    agent_names: Iterable[str],
    describe: bool,
    search: Optional[str],
) -> List[Dict[str, Any]]:
    agent_factory = ResourceAgentFacadeFactory(runner, report_processor)
    search_lower = search.lower() if search else None
    agent_list = []
    for name in agent_names:
        if search_lower and search_lower not in name.lower():
            continue
        try:
            split_name = split_resource_agent_name(name)
            metadata = (
                agent_factory.facade_from_parsed_name(split_name).metadata
                if describe
                else name_to_void_metadata(split_name)
            )
            agent_list.append(_agent_metadata_to_dict(metadata, describe))
        except ResourceAgentError as e:
            report_processor.report(
                resource_agent_error_to_report_item(
                    e, ReportItemSeverity.warning()
                )
            )
    return agent_list


# TODO return a DTO
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent
    """
    runner = lib_env.cmd_runner()
    report_processor = lib_env.report_processor
    agent_factory = ResourceAgentFacadeFactory(runner, report_processor)
    try:
        found_name = (
            split_resource_agent_name(agent_name)
            if ":" in agent_name
            else find_one_resource_agent_by_type(
                runner, report_processor, agent_name
            )
        )
        return _agent_metadata_to_dict(
            agent_factory.facade_from_parsed_name(found_name).metadata,
            describe=True,
        )
    except ResourceAgentError as e:
        lib_env.report_processor.report(resource_agent_error_to_report_item(e))
        raise LibraryError() from e



from .types import FakeAgentName, OcfVersion, _FAKE_AGENT_STANDARD

OCF_1_0 = OcfVersion("1.0")
OCF_1_1 = OcfVersion("1.1")
SUPPORTED_OCF_VERSIONS = [OCF_1_0, OCF_1_1]

FAKE_AGENT_STANDARD = _FAKE_AGENT_STANDARD
PACEMAKER_FENCED = FakeAgentName("pacemaker-fenced")

STONITH_ACTION_REPLACED_BY = ["pcmk_off_action", "pcmk_reboot_action"]

# OCF 1.0 doesn't define unique groups, they are defined since OCF 1.1. Pcs
# transforms OCF 1.0 agents to OCF 1.1 structure and therefore needs to create
# a group name for OCF 1.0 unique attrs. The name is: {this_prefix}{attr_name}
DEFAULT_UNIQUE_GROUP_PREFIX = "_pcs_unique_group_"



from typing import cast, List

from lxml.etree import _Element, SubElement

TAG = "group"


def is_group(resource_el: _Element) -> bool:
    return resource_el.tag == TAG


def append_new(resources_section: _Element, group_id: str) -> _Element:
    return SubElement(resources_section, TAG, id=group_id)


def get_inner_resources(
    group_el: _Element,
) -> List[_Element]:
    return cast(List[_Element], group_el.xpath("./primitive"))



from typing import Any, Dict, List, Optional

from pcs.lib.commands.resource_agent import (
    _agent_metadata_to_dict,
    _complete_agent_list,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.resource_agent import (
    InvalidResourceAgentName,
    list_resource_agents,
    ResourceAgentError,
    resource_agent_error_to_report_item,
    ResourceAgentFacadeFactory,
    ResourceAgentName,
)


# TODO return a list of DTOs
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all stonith agents on the local host, optionally filtered and described

    describe -- load and return agents' description as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()
    std_prov = "stonith"
    agent_names = [
        f"{std_prov}:{agent}"
        for agent in list_resource_agents(runner, std_prov)
    ]
    return _complete_agent_list(
        runner, lib_env.report_processor, agent_names, describe, search
    )


# TODO return a DTO
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent (not containing "stonith:" prefix)
    """
    runner = lib_env.cmd_runner()
    agent_factory = ResourceAgentFacadeFactory(runner, lib_env.report_processor)
    try:
        if ":" in agent_name:
            raise InvalidResourceAgentName(agent_name)
        return _agent_metadata_to_dict(
            agent_factory.facade_from_parsed_name(
                ResourceAgentName("stonith", None, agent_name)
            ).metadata,
            describe=True,
        )
    except ResourceAgentError as e:
        lib_env.report_processor.report(
            resource_agent_error_to_report_item(e, is_stonith=True)
        )
        raise LibraryError() from e
